import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Pakkaa on LZC-tiedostojen pakkaamiseen tarkoitettu luokka.
 */
public class Pakkaa {
	private static final int edistymisVali = 1000000;		// edistymispisteiden kirjoitusvli
	private static final int oletusKoodipituus = 19;
	private static final int oletusTeholaskuRaja = 5;		// suurin sallittu tehon lasku (%)

	/**
	 * Main-metodi ksittelee kynnistysparametrit ja kutsuu tiedoston pakkaajaa.
	 * @param args			Kynnistysparametrit.
	 */
	public static void main(String[] args) {
		if( args.length == 0 ) {					// ei parametreja? => nytetn ohje
			naytaOhje();
			return;
		}
		
		String lahdeNimi = null;					// lhde- ja
		String kohdeNimi = null;					// kohdetiedostojen nimet
		int tiedotusTaso = 0;						// kuinka paljon tietoa tulostetaan
		boolean tilastotulostus = false;			// tulostetaanko tilastoja varten
		int koodipituus = oletusKoodipituus;		// suurin kytettv koodipituus
		int teholaskuRaja = oletusTeholaskuRaja;	// suurin sallittu tehon lasku prosenteissa
		boolean patkatarkkailu = false;				// tarkkaillaanko pakkaustehoa ptkill
		
		// ksitelln kynnistysparametrit
		for( int i = 0; i < args.length; ++i ) {
			if( args[i].charAt(0) == '-' ) {		// parametri on valitsin
				String valitsin = args[i].substring(1);
				if( valitsin.equals("h") ) {				// -h    => nytetn ohje
					naytaOhje();
					return;
				} else if( valitsin.equals("a") ) {			// -a    => ei kooditaulun nollausta
					teholaskuRaja = -1;
				} else if( valitsin.equals("p") ) {			// -p    => ptktarkkailu kytss
					patkatarkkailu = true;
				} else if( valitsin.equals("v") ) {			// -v    => suurempi tiedotustaso
					++tiedotusTaso;
				} else if( valitsin.equals("s") ) {			// -s    => tilastotulostus
					tilastotulostus = true;
				} else if( valitsin.equals("n") ) {			// -n <pituus>    => asetetaan max.
					++i;									//                   koodipituus
					if( i == args.length ) {						// parametrit loppu?
						System.out.println("Virheelliset parametrit: Pituus puuttuu.");
						return;
					}
					try {
						koodipituus = Integer.parseInt(args[i]);
						if( koodipituus < 9 || koodipituus > 24 )
							throw new NumberFormatException();
					} catch( NumberFormatException e) {			// huono koodipituus?
						System.out.println("Virheelliset parametrit: Koodipituuden tytyy" +
								" olla kokonaisluku vlilt 9-24.");
						return;
					}
				} else if( valitsin.equals("o") ) {			// -o <tiedosto>    => kohdetiedos-
					++i;									//                     ton nimi
					if( i == args.length ) {						// parametrit loppu?
						System.out.println("Virheelliset parametrit: Kohdetiedoston nimi" +
								" puuttuu.");
						return;
					}
					kohdeNimi = args[i];
				} else if( valitsin.equals("r") ) {			// -r <raja>    => asetetaan
					++i;									//                 teholaskuraja
					if( i == args.length ) {						// parametrit loppu?
						System.out.println("Virheelliset parametrit: Rajasuhde puuttuu.");
						return;
					}
					try {
						teholaskuRaja = Integer.parseInt(args[i]);
						if( teholaskuRaja < 1 || teholaskuRaja > 100 )
							throw new NumberFormatException();
					} catch( NumberFormatException e) {			// huono rajasuhde?
						System.out.println("Virheelliset parametrit: Rajasuhteen tytyy" +
								" olla kokonaisluku vlilt 1-100.");
						return;
					}
				} else {									// tuntematon kynnistysparametri
					System.out.println("Virheelliset parametrit: Tuntematon valitsin -" +
							valitsin);
					return;
				}
			} else					// parametri on pakattavan tiedoston nimi
				if( lahdeNimi != null ) {		// nimi jo asetettu?
					System.out.println("Virheelliset parametrit: Useamman tiedoston" +
							" pakkaaminen ei onnistu");
					return;
				} else {
					lahdeNimi = args[i];
					if( kohdeNimi == null )	// jos kohdenime ei asetettu, keksitn se
						kohdeNimi = lahdeNimi + ".lzc";
				}
		}
		
		if( lahdeNimi == null ) {		// pakattava tiedosto asetettu?
			System.out.println("Virheelliset parametrit: Pakattavan tiedoston nimi puuttuu.");
			return;
		}
		
		// parametrit ok => suoritetaan pakkaus
		try {
			// tulostetaan infoa
			if( tiedotusTaso > 0 )
				System.out.println("Pakataan: " + lahdeNimi + " -> " + kohdeNimi);
			if( tiedotusTaso > 2 )
				if( teholaskuRaja < 0 )
					System.out.println("Kooditaulun pakkausaikainen alustus ei kytss.");
				else
					System.out.printf("Suurin sallittu pakkaustehon huonontuminen: %d %%\n",
							teholaskuRaja);
			
			// pakkaus
			long aloitusAika = System.currentTimeMillis();						// otetaan aikaa
			InputStream lahde = new BufferedInputStream(new FileInputStream(lahdeNimi));
			OutputStream kohde = new BufferedOutputStream(new FileOutputStream(kohdeNimi));
			double rajatehosuhde = (100 + teholaskuRaja) / 100.0;
			if( teholaskuRaja < 0 )
				rajatehosuhde = 0;
			String tilastotuloste = pakkaa(lahde, kohde, koodipituus, rajatehosuhde,
					patkatarkkailu, tiedotusTaso);		// pakataan
			lahde.close();
			kohde.close();
			long lopetusAika = System.currentTimeMillis();
			
			// tulostetaan lis infoa
			if( tiedotusTaso > 0 )
				System.out.println("Aikaa kului " + (lopetusAika-aloitusAika) / 1000.0 + " s");
			if( tilastotulostus )
				System.out.printf(" %10d; %10d; %10b;%s %10.3f;",
						koodipituus,
						teholaskuRaja,
						patkatarkkailu,
						tilastotuloste,
						(lopetusAika-aloitusAika) / 1000.0);
		} catch( java.io.FileNotFoundException e ) {		// ei tiedostoa?
			System.out.println("\nVirhe! Tiedostoa ei lydy: " + e.getLocalizedMessage());
		} catch( Exception e ) {							// virhe pakkauksessa?
			System.out.println("\nVirhe! Pakkausta ei voitu suorittaa: " +
					e.getLocalizedMessage());
		}
	}
	
	
	/**
	 * Tulostaa ruudulle ohjeen ohjelman kytst.
	 */
	private static void naytaOhje() {
		System.out.print(
			"Kytt: java Pakkaa [valitsimet] tiedosto\n" +
			"    Valitsimet:\n" +
			"      -a              l kyt kooditaulun alustusta kesken pakkauksen.\n" +
			"      -h              Nytt tmn ohjeen.\n" +
			"      -n pituus       Mritt suurimman kytettvn koodipituuden. Pituuden\n" +
			"                      tulee olla vlilt 9-24. Oletuspituus on " +
					oletusKoodipituus + ".\n" +
			"      -o tiedosto     Mritt kohdetiedoston nimen.\n" +
			"      -p              Pakkaustehoa tarkastellaan kaiken edellisen kooditaulun\n" +
			"                      alustuksen jlkeen pakatun datan sijaan edellisen\n" +
			"                      tarkastuksen jlkeen pakatun datan suhteen.\n" +
			"      -r suhde        Mritt suurimman sallitun pakkausuhteen huonontumisen\n"+
			"                      parhaasta edellisen alustamisen jlkeen saavutetusta\n" +
			"                      arvosta. Kun suhde ylittyy, alustetaan kooditaulu.\n" +
			"                      Annetaan prosentteina vlilt 1-100. Oletus on " +
					oletusTeholaskuRaja + ".\n" +
			"      -s              Tulostaa tietorivin tilastointia varten.\n" +
			"      -v              Nytt tietoa pakkauksen kulusta. Valitsimia voi antaa\n" +
			"                      useita, jolloin nytetn aina enemmn tietoa.\n");
	}
	
	
	/**
	 * Pakkaa annetun lhdetiedoston annettuu kohdetiedostoon annetulla koodipituudella.
	 * @param lahde			Pakattava lhdetiedosto.
	 * @param kohde			Pakattu kohdetiedosto.
	 * @param koodipituus		Suurin sallittu koodipituus.
	 * @param rajatehosuhde	Suurin sallittu nykyisen ja parhaan pakkaustehon suhde.
	 * @param tiedotusTaso		Mit suurempi, sit enemmn tulostetaan tietoa pakkauksesta.
	 * @return					Tilastoinnissa kytettvn yhteenvedon pakkauksesta.
	 */
	private static String pakkaa(InputStream lahde, OutputStream kohde,
			int koodipituus, double rajatehosuhde, boolean patkatarkkailu,
			int tiedotusTaso) throws IOException {
		PakattavaTiedosto paketti = new PakattavaTiedosto(koodipituus, kohde);
		PakkausTaulu kooditaulu = new PakkausTaulu(koodipituus);
		
		// tulostetaan mahdolliset alkutiedostukset
		if( tiedotusTaso > 2 ) {
			System.out.println("Koodipituus: " + koodipituus);
		}
		if( tiedotusTaso > 1 )
			System.out.print("Pakataan...");
		
		long kokoPakattu, kokoAlkuperainen = 0;	// tiedostojen koot tavuina
		int etenemislaskuri = 0;				// kytetn pakkauksen etenemis -tulostukseen
		int alustamisia = 0;					// montako kertaa kooditaulu alustettu
		
		int koodipituusNyt = kooditaulu.annaKoodipituus();	// kytettv koodipituus
		int edellinenKoodi = lahde.read();		// edellinen saatu koodi
		int uusiTavu;							// uusi luettu tavu
		int uusiKoodi;							// uusi saatu koodi
		
		long bittejaSisaan = 0;					// bittej luettu viimeisen alustuksen jlkeen
		long bittejaUlos = 0;					// bittej kirjoitettu          "
		long ensimmainenTarkastus = 9 << 9;	// ensimminen tehotarkastus
		long tarkastusVali = 9 << 9;			// tehotarkastusten vli
		long seuraavaTarkastus = ensimmainenTarkastus;	// seuraava bittejaUlos -mr,
														//   jolla suoritetaan tehotarkastus
		double aloitusParasTeho = 9.0 / 8.0;
		double parasTeho = aloitusParasTeho;	// paras saavutettu pakkausteho
		
		// kydn lpi jokainen lhdetiedoston tavu
		while( (uusiTavu = lahde.read()) != -1 ) {
			++kokoAlkuperainen;
			bittejaSisaan += 8;
			
			// haetaan edellist koodia + uutta tavua vastaava koodi
			koodipituusNyt = kooditaulu.annaKoodipituus();
			uusiKoodi = kooditaulu.haeKoodi(edellinenKoodi, uusiTavu);
			
			// jos vastaavaa koodia ei lytynyt, kirjoitetaan edellinen koodi tiedostoon
			if( uusiKoodi == -1 ) {
				paketti.kirjoita(edellinenKoodi, koodipituusNyt);
				bittejaUlos += koodipituusNyt;
				edellinenKoodi = uusiTavu;
				
				// tarkistetaan tmnhetkinen pakkausteho
				if( rajatehosuhde > 1.0 && bittejaUlos >= seuraavaTarkastus ) {
					double teho = (double)bittejaUlos / bittejaSisaan;
					if( teho < parasTeho )		// pivitetn parasTeho
						parasTeho = teho;
					//seuraavaTarkastus += koodipituusNyt << (koodipituusNyt-2);
					seuraavaTarkastus += tarkastusVali;
					// jos ptktarkkailu kytss, tyhjennetn arvot
					if( patkatarkkailu ) {
						bittejaSisaan = 0;
						bittejaUlos = 0;
					}
					// jos pakkausteho on laskenut liikaa, tyhjennetn kooditaulu
					if( teho / parasTeho > rajatehosuhde ) {
						kooditaulu.alusta();
						bittejaSisaan = 0;
						bittejaUlos = 0;
						parasTeho = aloitusParasTeho;
						paketti.kirjoita((1 << koodipituusNyt) - 1, koodipituusNyt);
						seuraavaTarkastus = ensimmainenTarkastus;
						++alustamisia;
					}
				}
			// koodi lytyi => ei tehd mitn
			} else {
				edellinenKoodi = uusiKoodi;
			}
			
			// tulostetaan tarvittaessa pakkauksen etenemis -piste
			if( tiedotusTaso > 1 ) {
				if( etenemislaskuri == 0 )
					System.out.print('.');
				etenemislaskuri = (etenemislaskuri + 1) % edistymisVali;
			}
		}
		
		// jo lhdetiedosto ei ollut tyhj, kirjoitetaan viimeisin saatu koodi tiedostoon
		if( edellinenKoodi >= 0 ) {
			paketti.kirjoita(edellinenKoodi, koodipituusNyt);
			++kokoAlkuperainen;
		}
		paketti.paata();		// ptetn paketti
		
		// tulostetaan mahdolliset lopputiedostukset
		kokoPakattu = paketti.paljonkoKirjoitettu();
		if( tiedotusTaso > 1 )
			System.out.println("OK");
		if( tiedotusTaso > 2 ) {
			System.out.printf("Hajautustauluhakuja keskimrin: %.3f\n",
					kooditaulu.hakujaKeskimaarin());
			System.out.println("Kooditaulun alustuksia: " + alustamisia);
		}
		if( tiedotusTaso > 1 ) {
			System.out.println("Alkuperinen koko: " + kokoAlkuperainen);
			System.out.println("Pakattu koko: " + kokoPakattu);
		}
		if( tiedotusTaso > 0 )
			if( kokoAlkuperainen > 0 )
				System.out.printf("Pakkausuhde: %.2f %%\n",
						100.0 * kokoPakattu / kokoAlkuperainen);
		
		int raja = (int)((rajatehosuhde - 1) * 100);
		return String.format(" %10.3f; %10d; %10d; %10d; %10.2f;",
				kooditaulu.hakujaKeskimaarin(),
				alustamisia,
				kokoAlkuperainen,
				kokoPakattu,
				100.0 * kokoPakattu / kokoAlkuperainen);
	}
}
