
/**
 * PakkausTaulu pit kirjaa kytettviss olevista koodeista ja niiden merkityksist.
 * PakkausTaulun avulla tavujonoja muutetaan koodeiksi.
 */
public class PakkausTaulu {
	// hajautustaulun koko koodipituuden mukaan (pituudet 9-24).
	private static final int[] tauluKoot = { 877, 1741, 3491, 6967, 13931, 27883,
			55711, 111427, 222823, 445649, 891311, 1782589, 3565169, 7130329,
			14260643, 28521293 };
	private final int maxKoodipituus;	// suurin kytettv koodipituus bittein
	private final int maxKoodi;		// suurin mahdollinen koodi
	private final int hashSiirto;		// hajatusarvon laskennassa kytettv arvo
	private final int taulunKoko;		// hajautustaulun koko
	private int koodipituus;			// hetkellisesti kytettv suurin mahd. koodipituus
	private int seuraavaVapaaKoodi;
	
	private int hakuja;				// taulusta tehtyjen hakujen lukumr
	private int jatkohakuja;		// kokelujonohakujen lkm
	
	// itse hajautustaulu, johon tallennetaan koodi ja sit vastaava koodi-tavu -yhdistelm
	private int[] koodi;
	private int[] edellinenKoodi;
	private int[] uusiTavu;
	
	/**
	 * Luo ja alustaa uuden hajautustaulun, kytten annettua maksimikoodipituutta.
	 * @param maxKoodipituus	Kuinka pitki koodeja tauluun tytyy saada voida list.
	 */
	public PakkausTaulu(int maxKoodipituus) {
		this.maxKoodipituus = maxKoodipituus;
		this.maxKoodi = (1 << this.maxKoodipituus) - 1;
		this.taulunKoko = tauluKoot[this.maxKoodipituus - 9];
		this.hashSiirto = this.maxKoodipituus - 8;
		this.koodi = new int[this.taulunKoko];
		this.edellinenKoodi = new int[this.taulunKoko];
		this.uusiTavu = new int[this.taulunKoko];
		this.hakuja = 0;
		this.jatkohakuja = 0;
		this.alusta();
	}
	
	/**
	 * Tyhjent hajautustaulun ja alustaa koodipituuden, jolloin pakkaus "alkaa alusta".
	 */
	public void alusta() {
		this.koodipituus = 9;
		for( int i = 0; i < this.taulunKoko; ++i )	// hajautustaulun alustus
			this.koodi[i] = -1;
		// koodit 0-255 on ennalta mritelty vastaamaan itsen vastaavaa tavua 
		this.seuraavaVapaaKoodi = 256;
	}
	
	/**
	 * Palauttaa parametreina annettavia edellist koodia sek siihen listtv tavua
	 * vastaavan uuden koodi. Jos koodia ei ole viel olemassa, lis sen tauluun ja
	 * palauttaa -1.
	 * @param edellinen		Edellinen haettu koodi.
	 * @param uusiTavu			Edellist koodia vastaavaan dataan listtv uusi tavu.
	 * @return					Sytett vastaavan koodiin, tai -1 jos koodia ei ole.
	 */
	public int haeKoodi(int edellinen, int uusiTavu) {
		int indeksi = edellinen ^ (uusiTavu << hashSiirto);	// lasketaan hajautusavain
		int indeksiLisays = (indeksi >> 1) + 1;				// ja siirto
		
		++this.hakuja;
		
		// etsitn koodi taulusta
		while( this.koodi[indeksi] != -1 ) {
			++this.jatkohakuja;
			if( this.edellinenKoodi[indeksi] == edellinen &&
					this.uusiTavu[indeksi] == uusiTavu )
				return this.koodi[indeksi];					// ja palautetaan se
			indeksi += indeksiLisays;
			if( indeksi >= this.taulunKoko )
				indeksi -= this.taulunKoko;
		}
		
		// jos koodia ei lytynyt ja tilaa on viel, listn koodi tauluun
		if( this.seuraavaVapaaKoodi < this.maxKoodi ) {
			if( this.seuraavaVapaaKoodi + 1 == 1 << this.koodipituus )
				++this.koodipituus;					// kasvatetaan tarvittaessa koodipituutta
			this.edellinenKoodi[indeksi] = edellinen;
			this.uusiTavu[indeksi] = uusiTavu;
			this.koodi[indeksi] = this.seuraavaVapaaKoodi;
			++this.seuraavaVapaaKoodi;
		}
		
		// jos koodia ei lytynyt, palautetaan -1
		return -1;
	}
	
	/**
	 * Kertoo seuraavan palautettavan koodin (suurimman mahdollisen) pituuden.
	 * @return					Koodipituuden bittein.
	 */
	public int annaKoodipituus() {
		return this.koodipituus;
	}
	
	/**
	 * Kertoo, montako hakua hajautustaulusta keskimrin piti tehd yht koodikysely kohden.
	 * @return					Keskimrisen hakujen mrn.
	 */
	public double hakujaKeskimaarin() {
		return (double)(this.hakuja + this.jatkohakuja) / this.hakuja;
	}
}
