import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Pura on LZC-tiedostojen purkamiseen tarkoitettu luokka.
 */
public class Pura {
	private static final int edistymisVali = 1000000;		// edistymispisteiden kirjoitusvli
	
	/**
	 * Main-metodi ksittelee kynnistysparametrit ja kutsuu tiedoston pakkaajaa.
	 * @param args			Kynnistysparametrit.
	 */
	public static void main(String[] args) {
		if( args.length == 0 ) {					// ei parametreja? => nytetn ohje
			naytaOhje();
			return;
		}
		
		String lahdeNimi = null;					// lhdetiedoston nimi
		String kohdeNimi = null;					// kohdetiedoston nimi
		int tiedotusTaso = 0;						// kuinka paljon tietoa tulostetaan
		boolean tilastotulostus = false;			// tulostetaanko tilastoja varten
		
		// ksitelln kynnistysparametrit
		for( int i = 0; i < args.length; ++i ) {
			if( args[i].charAt(0) == '-' ) {		// parametri on valitsin
				String valitsin = args[i].substring(1);
				if( valitsin.equals("h") ) {				// -h	 => nytetn ohje
					naytaOhje();
					return;
				} else if( valitsin.equals("v") ) {			// -v    => suurempi tiedotustaso
					++tiedotusTaso;
				} else if( valitsin.equals("s") ) {			// -s    => tilastotulostus
					tilastotulostus = true;
				} else if( valitsin.equals("o") ) {			// -o <tiedosto>    => kohdetiedos-
					++i;									//                     ton nimi
					if( i == args.length ) {
						System.out.println("Virheelliset parametrit: Kohdetiedoston nimi" +
								" puuttuu.");
						return;
					}
					kohdeNimi = args[i];
				} else {									// tuntematon kynnistysparametri
					System.out.println("Virheelliset parametrit: Tuntematon valitsin -" +
							valitsin);
					return;
				}
			} else									// parametri on purettavan tiedoston nimi
				if( lahdeNimi != null ) {
					System.out.println("Virheelliset parametrit: Useamman tiedoston" +
							" purkaminen ei onnistu");
					return;
				} else {
					lahdeNimi = args[i];
					if( kohdeNimi == null &&
							lahdeNimi.substring(lahdeNimi.length()-4,
								lahdeNimi.length()).equals(".lzc") )
						kohdeNimi = lahdeNimi.substring(0, lahdeNimi.length()-4);
				}
		}
		
		if( lahdeNimi == null ) {		// purettava tiedosto asetettu?
			System.out.println("Virheelliset parametrit: Purettavan tiedoston nimi puuttuu.");
			return;
		}
		
		if( kohdeNimi == null ) {		// kohdetiedostoa ei annettu eik pystytty pttelemn?
			System.out.println("Virheelliset parametrit: Mys kohdetiedoston nimi tytyy" +
					" antaa tss tapauksessa. (Nime ei pystytty pttelemn.)");
			return;
		}
		
		// parametrit ok => suoritetaan purku
		try {
			// tulostetaan infoa
			if( tiedotusTaso > 0 )
				System.out.println("Puretaan: " + lahdeNimi + " -> " + kohdeNimi);
			
			// purku
			long aloitusAika = System.currentTimeMillis();
			InputStream lahde = new BufferedInputStream(new FileInputStream(lahdeNimi));
			OutputStream kohde = new BufferedOutputStream(new FileOutputStream(kohdeNimi));
			pura(lahde, kohde, tiedotusTaso);
			lahde.close();
			kohde.close();
			long lopetusAika = System.currentTimeMillis();
			
			// tulostetaan lis infoa
			if( tiedotusTaso > 0 )
				System.out.println("Aikaa kului " + (lopetusAika-aloitusAika) / 1000.0 + " s");
			if( tilastotulostus )
				System.out.printf(" %10.3f;", (lopetusAika-aloitusAika) / 1000.0);
		} catch( java.io.FileNotFoundException e ) {		// ei tiedostoa?
			System.out.println("\nVirhe! Tiedostoa ei lydy: " + e.getLocalizedMessage());
		} catch( Exception e ) {							// virhe pakkauksessa?
			System.out.println("\nVirhe! Purkua ei voitu suorittaa: " +
					e.getLocalizedMessage());
		}
	}
	
	/**
	 * Tulostaa ruudulle ohjeen ohjelman kytst.
	 */
	private static void naytaOhje() {
		System.out.println(
			"Kytt: java Pura [valitsimet] tiedosto\n" +
			"    Valitsimet:\n" +
			"      -h              Nytt tmn ohjeen.\n" +
			"      -v              Nytt tietoa purkamisen kulusta. Valitsimia voi antaa\n" +
			"                      useita, jolloin nytetn aina enemmn tietoa.\n" +
			"      -s              Tulostaa tietorivin tilastointia varten.\n" +
			"      -o tiedosto     Mritt kohdetiedoston nimen.");
	}
	
	/**
	 * Pakkaa annetun lhdetiedoston annettuu kohdetiedostoon annetulla koodipituudella.
	 * @param lahde			Purettava lhdetiedosto.
	 * @param kohde			Purettu kohdetiedosto.
	 * @param tiedotusTaso		Mit suurempi, sit enemmn tulostetaan tietoa pakkauksesta.
	 */
	private static void pura(InputStream lahde, OutputStream kohde,
			int tiedotusTaso) throws IOException {
		PurettavaTiedosto paketti = new PurettavaTiedosto(lahde);
		int koodipituus = paketti.annaMaxKoodipituus();	// kysytn kytetty koodipituus
		
		PurkuTaulu kooditaulu = new PurkuTaulu(koodipituus);
		int[] data = new int[1 << koodipituus];			// puskuri, johon purettu data laitetaan
		int kokoPurettu = 0;							// puretun tiedoston koko tavuina
		 
		// tulostetaan mahdolliset alkutiedostukset
		if( tiedotusTaso > 2 )
			System.out.println("Koodipituus: " + koodipituus);
		if( tiedotusTaso > 1 )
			System.out.print("Puretaan...");
		
		int etenemislaskuri = 0;				// kytetn pakkauksen etenemis -tulostukseen
		int koodipituusNyt = kooditaulu.annaKoodipituus();		// hetkellinen koodipituus
		int koodi;								// tiedostosta luettu koodi
		int pituus;								// purettua koodia vastaavan datan mr
		
		// kydn lpi jokainen lhdetiedoston koodi
		while( (koodi = paketti.lue(koodipituusNyt)) != -1 ) {
			if( koodi == (1 << koodipituusNyt) - 1 ) {		// kooditaulun alustus -koodi?
				kooditaulu.alusta();
			} else {
				// puretaan koodi ja kirjoitetaan data kohdetiedostoon
				pituus = kooditaulu.haeData(koodi, data);
				while( pituus > 0 ) {
					--pituus;
					kohde.write(data[pituus]);
					++kokoPurettu;
				}
			}
			
			// tulostetaan tarvittaessa pakkauksen etenemis -piste
			if( tiedotusTaso > 1 ) {
				if( etenemislaskuri == 0 )
					System.out.print('.');
				etenemislaskuri = (etenemislaskuri + 1) % edistymisVali;
			}
			
			koodipituusNyt = kooditaulu.annaKoodipituus();
		}
		
		// tulostetaan mahdolliset lopputiedostukset
		if( tiedotusTaso > 0 )
			System.out.println("OK");
		if( tiedotusTaso > 1 )
			System.out.println("Puretun tiedoston koko: " + kokoPurettu);
	}
}
