import java.io.InputStream;
import java.io.IOException;

/**
 * PurettavaTiedosto on purettavan tiedoston lukemiseen tarkoitettu luokka, jonka
 * avulla tiedostosta saa luettua halutun mittaisia (bittein) koodeja.
 */
public class PurettavaTiedosto {
	private final int maxKoodipituus;			// suurin tiedostossa kytettv koodipituus
	private int puskuri;						// puskurissa silytetn viel kirjoittamatto-
	private int puskurinKoko;					//   mia bittej
	private InputStream tiedosto;				// luettava tiedosto
	
	/**
	 * Luo uuden PurettavaTiedosto:n, joka kytt annettua InputStreamia.
	 * @param tiedosto			Itse tiedosto, josta data puretaan.
	 */
	public PurettavaTiedosto(InputStream tiedosto) throws IOException {
		this.tiedosto = tiedosto;
		this.puskuri = 0;
		this.puskurinKoko = 0;
		this.maxKoodipituus = this.tiedosto.read();	// koodipituus luetaan tiedostosta
		if( this.maxKoodipituus == -1 )				// tiedosto on tyhj?
			throw new IOException("Ei voi purkaa. Tiedosto on korruptoitunut." +
					" (tyhj tiedosto)");
		if( this.maxKoodipituus < 9  || this.maxKoodipituus > 24 )	// virheellinen koodipituus?
			throw new IOException("Ei voi purkaa. Tiedosto on korruptoitunut." +
					" (vr koodipituus)");
	}
	
	/**
	 * Kertoo pakatussa tiedostossa kytetyn koodipituuden.
	 * @return					Koodipituuden bittein.
	 */
	public int annaMaxKoodipituus() {
		return this.maxKoodipituus;
	}
	
	/**
	 * Lukee tiedostosta annetun koodipituuden mittasen koodin ja palauttaa sen. Jos
	 * tiedosto on loppu, palauttaa -1.
	 * @param koodipituus		Luettavan koodin koko bittein.
	 * @return					Luetun koodin, tai -1 jos tiedosto on loppu.
	 */
	public int lue(int koodipituus) throws IOException {
		int luettu = this.tiedosto.read();
		if( luettu == -1 )								// tiedosto loppu (ei koodeja jljell)?
			return -1;
		
		// tytetn puskuria tiedostosta niin kauan, ett saavutetaan tarvittava mr
		// (eli koodipituuden verran) bittej
		this.puskuri <<= 8;
		this.puskuri |= luettu;
		this.puskurinKoko += 8;
		while( this.puskurinKoko < koodipituus ) {
			int uusiTavu = this.tiedosto.read();
			if( uusiTavu == -1 )						// tiedosto loppui kesken koodin?
				throw new IOException("Ei voi purkaa. Tiedosto on korruptoitunut." +
						" (tiedosto loppuu ylltten");
			this.puskuri <<= 8;
			this.puskuri |= uusiTavu;
			this.puskurinKoko += 8;
		}
		
		// luetaan koodi puskurista ja palautetaan se
		this.puskurinKoko -= koodipituus;
		int koodi = this.puskuri >>> this.puskurinKoko;
		this.puskuri &= (1 << this.puskurinKoko) - 1;		// nollataan koodin varaama tila
		return koodi;
	}
}
