import java.io.IOException;

/**
 * PurkuTaulu pit kirjaa kytettviss olevista koodeista ja niiden merkityksist.
 */
public class PurkuTaulu {
	private int maxKoodipituus;			// suurin kytettv koodipituus
	private int koodipituus;				// hetkellisesti kytettv suurin mahd. koodipituus
	private int taulunKoko;					// purkamisessa kytettvienn taulukoiden koko
	private int seuraavaVapaaKoodi;			// pienin koodi, jolla ei ole viel merkityst
	private int edellinenKysyttyKoodi;
	private int edellisenEnsimmainenTavu;	// edellisen palautetun dataptkn ensimminen tavu
	
	// taulukot koodeja (indeksi) vastaavista koodi-tavu -yhdistelmist
	private int[] edellinenKoodi;
	private int[] uusiTavu;
	
	/**
	 * Luo ja alustaa PurkuTaulun annetulla koodipituudella.
	 * @param maxKoodipituus	Kuinka pitki koodeja taulun tytyy voida sislt.
	 */
	public PurkuTaulu(int maxKoodipituus) {
		this.maxKoodipituus = maxKoodipituus;
		this.taulunKoko = (1 << this.maxKoodipituus) - 1;		// = 2^maxKoodipituus
		this.edellinenKoodi = new int[this.taulunKoko];
		this.uusiTavu = new int[this.taulunKoko];
		this.alusta();
	}
	
	/**
	 * Tyhjent sanakirjan ja alustaa koodipituuden.
	 */
	public void alusta() {
		this.koodipituus = 9;
		this.seuraavaVapaaKoodi = 256;					// koodit 0-255 jo valmiiksi mritelty
		this.edellinenKysyttyKoodi = -1;				// yhtn ei viel kysytty!
		this.edellisenEnsimmainenTavu = -1;
	}
	
	/**
	 * Sijoittaa annettua koodia vastaavan dataptkn KNTEISESS jrjestyksess,
	 * parametrina annettuun taulukkoon. Paluuarvo on palautetun datan pituus.
	 * @param koodi			Purettava koodi.
	 * @param data				Taulukko, jossa purettu data palautetaan.
	 * @return					data-taulukossa palautettavan datan mrn tavuina.
	 */
	public int haeData(int koodi, int[] data) throws IOException {
		int pituus = 0;					// dataptkn pituus
		int uusiKysyttyKoodi = koodi;
		// Erikoistapaus: Jos koodia ei viel ole hakemistossa, saadaan se edellisest
		// kysytyst koodista liittmll loppuun (eli alkuun ;) sit vastaavan
		// dataptkn ensimminen tavu.
		if( koodi == this.seuraavaVapaaKoodi ) {
			data[pituus] = this.edellisenEnsimmainenTavu;
			++pituus;
			koodi = this.edellinenKysyttyKoodi;
		} else if( koodi > this.seuraavaVapaaKoodi ) {
			throw new IOException("Data on korruptoitunut. (liian iso koodi: " +
					koodi + "/" + this.seuraavaVapaaKoodi + ")");
		}
		// haetaan koodia vastaava dataptk taulusta
		while( koodi >= 256 ) {
			data[pituus] = this.uusiTavu[koodi];
			++pituus;
			koodi = this.edellinenKoodi[koodi];
		}
		data[pituus] = koodi;					// koodit 0-255 vastaavat tavuja 0-255
		++pituus;
		
		// listn tauluun uusi koodi (= edellinen koodi + nykyisen datan ensimminen tavu)
		if( this.edellinenKysyttyKoodi != -1 && this.seuraavaVapaaKoodi < this.taulunKoko ) {
			this.edellinenKoodi[this.seuraavaVapaaKoodi] = this.edellinenKysyttyKoodi;
			this.uusiTavu[this.seuraavaVapaaKoodi] = koodi;
			++this.seuraavaVapaaKoodi;
			if( this.seuraavaVapaaKoodi + 1 == 1 << this.koodipituus &&
					this.koodipituus < this.maxKoodipituus )
				++this.koodipituus;				// kasvatetaan tarvittaessa koodipituutta
		}
		
		this.edellisenEnsimmainenTavu = koodi;
		this.edellinenKysyttyKoodi = uusiKysyttyKoodi;
		
		// palautetaan dataptkn pituus
		return pituus;
	}
	
	/**
	 * Kertoo seuraavan koodin (suurimman mahdollisen) pituuden.
	 * @return					Koodipituuden bittein.
	 */
	public int annaKoodipituus() {
		return this.koodipituus;
	}
}
